package main

import (
	"bytes"
	"fmt"
	"go/ast"
	"go/format"
	"go/parser"
	"go/token"
	"os"
	"strings"
	"text/template"
)

func errx(msg string, args ...interface{}) {
	fmt.Fprintf(os.Stderr, msg, args...)
	os.Exit(1)
}

// enjargo me
type Argo struct {
	Num  int // skip
	name string
	hide string // enjargo:skip
	Data struct {
		Type  string
		Value string
	}
	bananas []string
	Last    int
}

func main() {
	var fset token.FileSet
	pkgs, err := parser.ParseDir(&fset, ".", nil, parser.ParseComments)
	if err != nil {
		errx("error parsing packages: %s", err)
	}
	for _, pkg := range pkgs {
		process(pkg)
	}
	a := Argo{
		Num:     42,
		name:    "howdy",
		hide:    "alert",
		bananas: []string{"a", "b", "c"},
	}
	a.Data.Type = "prof"
	a.Data.Value = "spaceman"
	a.WriteJson(os.Stdout) // won't compile until the first time loop is completed
}

var skel = `package main

import (
	"fmt"
	"io"
)

{{ define "field" }}
{{ if .Fields }}
{{ $first := true }}
	fmt.Fprintf(w, "{{ .Indent }}\"{{ .JsonName }}\": {\n")
{{ range .Fields }}
{{ if $first }}{{ $first = false }}{{ else }}fmt.Fprintf(w, ",\n"){{ end }}
{{ template "field" . }}{{ end }}
fmt.Fprintf(w, "\n{{ .Indent }}}")
{{ else }}
{{ if .Range }}
fmt.Fprintf(w, "{{ .Indent}}\"{{ .JsonName }}\": [\n")
for i := range x.{{ .Name }} {
	if i > 0 { fmt.Fprintf(w, ",\n") }
	fmt.Fprintf(w, "  {{ .Indent}} {{ .Fmt }}", x.{{ .Name }}[i])
}
fmt.Fprintf(w, "\n{{ .Indent }}]")
{{ else }}
	fmt.Fprintf(w, "{{ .Indent}}\"{{ .JsonName }}\": {{ .Fmt }}", x.{{ .Name }})
{{ end }}
{{ end }}
{{ end }}
{{ range . }}
func (x *{{ .Name }}) WriteJson(w io.Writer) {
	fmt.Fprintf(w, "{\n")
{{ $first := true }}
{{ range .Fields }}
{{ if $first }}{{ $first = false }}{{ else }}fmt.Fprintf(w, ",\n"){{ end }}
{{ template "field" . }}
{{ end }}
	fmt.Fprintf(w, "\n}\n")
}
{{ end }}
`

var templ = template.Must(template.New("skel").Parse(skel))

// enjargo
type EnType struct {
	Name   string
	Fields []*EnField
}

// info for a field to be encoded
type EnField struct {
	Indent   string
	JsonName string
	Fields   []*EnField
	Name     string
	Fmt      string
	Range    bool
}

func formatstring(ident string) string {
	switch ident {
	case "int", "int64":
		return `%d`
	case "bool":
		return `%v`
	case "string":
		return `\"%s\"`
	default:
		return ""
	}
}

func collectFields(decl *ast.GenDecl) *EnType {
	skip := true
	if comm := decl.Doc; comm != nil {
		for _, c := range comm.List {
			if strings.Contains(c.Text, "enjargo") {
				skip = false
				break
			}
		}
	}
	if skip {
		return nil
	}
	prefix := ""
	indent := "  "
	var dumpstruct func(*ast.StructType) []*EnField
	dumpstruct = func(str *ast.StructType) []*EnField {
		dumpsimple := func(name string, ident *ast.Ident) *EnField {
			ef := new(EnField)
			ef.Indent = indent
			ef.JsonName = name
			ef.Name = prefix + name
			ef.Fmt = formatstring(ident.Name)
			return ef
		}
		dumparray := func(name string, arr *ast.ArrayType) *EnField {
			ident, ok := arr.Elt.(*ast.Ident)
			if !ok {
				return nil
			}
			ef := new(EnField)
			ef.Range = true
			ef.Indent = indent
			ef.JsonName = name
			ef.Name = prefix + name
			ef.Fmt = formatstring(ident.Name)
			return ef
		}
		var efs []*EnField
		for _, field := range str.Fields.List {
			skip := false
			if comm := field.Comment; comm != nil {
				for _, c := range comm.List {
					if strings.Contains(c.Text, "enjargo:skip") {
						skip = true
						break
					}
				}
			}
			if skip {
				continue
			}
			name := field.Names[0].Name
			ident, ok := field.Type.(*ast.Ident)
			if ok {
				ef := dumpsimple(name, ident)
				if ef != nil && ef.Fmt != "" {
					efs = append(efs, ef)
				}
				continue
			}
			arr, ok := field.Type.(*ast.ArrayType)
			if ok {
				ef := dumparray(name, arr)
				if ef != nil && ef.Fmt != "" {
					efs = append(efs, ef)
				}
				continue
			}
			str, ok := field.Type.(*ast.StructType)
			if ok {
				ef := new(EnField)
				ef.Indent = indent
				ef.JsonName = name
				ef.Name = prefix + name
				old := prefix
				prefix = prefix + name + "."
				indent = indent + "  "
				ef.Fields = dumpstruct(str)
				indent = indent[:len(indent)-2]
				prefix = old
				efs = append(efs, ef)
				continue
			}
			continue
		}
		return efs
	}
	spec, ok := decl.Specs[0].(*ast.TypeSpec)
	if !ok {
		fmt.Printf("no typesec in decl\n")
		return nil
	}
	name := spec.Name.Name
	str, ok := spec.Type.(*ast.StructType)
	if !ok {
		fmt.Printf("no struct in %s\n", name)
		return nil
	}
	entype := new(EnType)
	entype.Name = name
	entype.Fields = dumpstruct(str)
	return entype
}

func process(pkg *ast.Package) {
	var entypes []*EnType
	walker := func(node ast.Node) bool {
		decl, ok := node.(*ast.GenDecl)
		if !ok || decl.Tok != token.TYPE {
			return true
		}
		et := collectFields(decl)
		if et != nil {
			entypes = append(entypes, et)
		}
		return true
	}
	for _, file := range pkg.Files {
		ast.Inspect(file, walker)
	}
	var buf bytes.Buffer
	templ.Execute(&buf, entypes)
	out, err := format.Source(buf.Bytes())
	if err != nil {
		errx("can't format result: %s", err)
	}
	w, err := os.Create("enjargo_methods.go")
	if err != nil {
		errx("can't open output: %s", err)
	}
	defer w.Close()
	w.Write(out)
}
